/* Version $Id: TLGChannelBox.cc 6485 2011-09-09 16:21:47Z james.batch@LIGO.ORG $ */
/* -*- mode: c++; c-basic-offset: 3; -*- */
/*---------------------------------------------------------------------------*/
/*                                                                           */
/* Module Name:  TLGChannelBox						     */
/*                                                                           */
/* Module Description:  ROOT GUI widget for handling LIGO channel names      */
/*                                                                           */
/*---------------------------------------------------------------------------*/

#include "ligogui/TLGChannelBox.hh"
#include <stdlib.h>
#include <cstring>
#include <ctype.h>
#include <iostream>
#include <TVirtualX.h>
#include <TGClient.h>
#include <TGPicture.h>
#include <TSystem.h>

namespace ligogui {
   using namespace std;


//--- Some utility functions ---------------------------------------------------
#if 0
   static Int_t FontHeight(FontStruct_t f)
   {
      int max_ascent, max_descent;
      gVirtualX->GetFontProperties(f, max_ascent, max_descent);
      return max_ascent + max_descent;
   }

   static Int_t FontAscent(FontStruct_t f)
   {
      int max_ascent, max_descent;
      gVirtualX->GetFontProperties(f, max_ascent, max_descent);
      return max_ascent;
   }

   static Int_t FontTextWidth(FontStruct_t f, const char *c)
   {
      return gVirtualX->TextWidth(f, c, strlen(c));
   }
#endif


//////////////////////////////////////////////////////////////////////////
//                                                                      //
// ChannelTree                                                          //
//                                                                      //
// A list of channel names                                              //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

   ChannelTree::ChannelTree (const ChannelEntry* chns, UInt_t chnnum,
                     Int_t type)
      : fChnClient (0), fLeftArrow (0), fChannels (0), fChnNum (0), 
	fChannelTreeType (type)
   {
      SetChannels (chns, chnnum);
   }


   ChannelTree::ChannelTree (const char* const* chnnames, UInt_t chnnum,
                     Int_t type)
      : fChnClient (0), fLeftArrow (0), fChannels (0), fChnNum (0), 
	fChannelTreeType (type)
   {
      SetChannels (chnnames, chnnum);
   }


   ChannelTree::ChannelTree (const char* chnnames, Int_t type)
      : fChnClient (0), fLeftArrow (0), fChannels (0), fChnNum (0), 
	fChannelTreeType (type)
   {
      if (chnnames != 0) {
         SetChannels (chnnames);
      }
   }


   ChannelTree::~ChannelTree ()
   {
      if (fChannels != 0) {
         delete [] fChannels;
         fChannels = 0;
         fChnNum = 0;
      }
      if (fChnClient && fLeftArrow) {
	 fChnClient->FreePicture(fLeftArrow);
	 fLeftArrow = 0;
      }      
   }

   void  
   ChannelTree::SetChannelClient(TGClient* tc) {
      if (fChnClient && fLeftArrow) {
	 fChnClient->FreePicture(fLeftArrow);
	 fLeftArrow = 0;
      }
      fChnClient = tc;
      fLeftArrow = fChnClient->GetPicture("arrow_left.xpm");
   }

   Bool_t ChannelTree::SetChannels (const ChannelEntry* chns, 
                     UInt_t chnnum)
   {
      // allocate memory
      if (chnnum < 0) {
         return kFALSE;
      }
      if (fChannels != 0) {
         delete [] fChannels;
         fChannels = 0;
         fChnNum = 0;
      }
      if ((chnnum == 0) || (chns == 0)) {
         return kTRUE;
      }
      fChannels = new ChannelEntry [chnnum];
      if (fChannels == 0) {
         return kFALSE;
      }
      fChnNum = chnnum;
      for (UInt_t i = 0; i < fChnNum; ++i) {
         fChannels[i] = chns[i];
      //memcpy (fChannels, chns, fChnNum * sizeof (ChannelEntry));
      }
      return kTRUE;
   }


   Bool_t ChannelTree::SetChannels (const char* const* chnnames, 
                     UInt_t chnnum)
   {
      // allocate memory
      if (chnnum < 0) {
         return kFALSE;
      }
      if (fChannels != 0) {
         delete [] fChannels;
         fChannels = 0;
         fChnNum = 0;
      }
      if ((chnnum == 0) || (chnnames == 0)) {
         return kTRUE;
      }
      fChannels = new ChannelEntry [chnnum];
      if (fChannels == 0) {
         return kFALSE;
      }
      fChnNum = chnnum;
   
      // assign channels
      for (UInt_t i = 0; i < chnnum; i++) {
         fChannels[i].SetName (chnnames[i]);
         fChannels[i].SetUDN ("");
         fChannels[i].SetRate (0);
      }
      return kTRUE;
   }


   Bool_t ChannelTree::SetChannels (const char* chnnames)
   {
      return MakeChannelList (chnnames, fChannels, fChnNum);
   }


   Bool_t ChannelTree::MakeChannelList (const char* chnnames,
                     ChannelEntry*& chns, UInt_t& chnnum)
   {
      // allocate memory
      if (chns != 0) {
         delete [] chns;
         chns = 0;
         chnnum = 0;
      }
      if (chnnames == 0) {
         return kTRUE;
      }
      UInt_t size = 1;
      int i = 0;
      while  (chnnames[i] != 0) {
         if (isspace (chnnames[i])) {
            size++;
         }
         i++;
      }
      chns = new ChannelEntry [size];
      if (chns == 0) {
         return kFALSE;
      }
      // assign channel names
      char* p = new char [strlen (chnnames) + 1];
      if (p == 0) {
         return kFALSE;
      }
      strcpy (p, chnnames);
      char* lasts;
      char* tok = strtok_r (p, " \t\n\f\r\v", &lasts);
      while ((tok != 0) && (chnnum < size)) {
         chns[chnnum].SetName (tok);
         chns[chnnum].SetUDN ("");
         chns[chnnum].SetRate (0);
         tok = strtok_r (0, " \t\n\f\r\v", &lasts);
         // check if following token represents a UDN
         if (tok && (*tok == '@')) {
            chns[chnnum].SetUDN (tok + 1);
            tok = strtok_r (0, " \t\n\f\r\v", &lasts);
         }
         // check if following token represents sampling rate?
         bool num = true;
         for (char* c = tok; c && *c && num; c++) {
            num = isdigit (*c) || (*c == '.');
         }
         if (tok && *tok && num) {
            chns[chnnum].SetRate (atof (tok));
            tok = strtok_r (0, " \t\n\f\r\v", &lasts);
         }
         chnnum++;
      }
   
      delete [] p;
      return kTRUE;
   }


   Bool_t ChannelTree::GetIfoSub (const char* chn, 
                     char* ifo, char* sub, char* rest)
   {
      *ifo = 0;
      *sub = 0;

      const char* cur = chn;
      const char* p = strchr (cur, ':');
      if (p == 0) {
         return kFALSE;
      }
      size_t lifo = p - cur;
      if (lifo > 63) lifo = 63;
      memcpy(ifo, cur, lifo);
      ifo[lifo] = 0;
      cur = p+1;

      // vacuum channels!
      char* w = strchr (ifo, '-');
      if (w) {
	 size_t lsub = lifo;
	 lifo = w - ifo;
         *w++ = 0;

	 lsub -= lifo + 1;
         memcpy (sub, w, lsub);
         sub[lsub] = 0;

	 size_t lrest = strlen(cur);
	 if (lrest > 255) lrest = 255;
         memcpy(rest, cur, lrest);
         rest[lrest] = 0;
         return  (lifo == 3 && lsub == 2) ? kTRUE : kFALSE;
      }
      // normal channels
      else {
         p = strchr (cur, '-');
         if (p == NULL) {
            return kFALSE;
         }
         
	 size_t lsub = p - cur;
         memcpy (sub, cur, lsub);
         sub[lsub] = 0;

	 size_t lrest = strlen(++p);
         strncpy (rest, p, lrest);
         rest[lrest] = 0;
         return (lifo == 2 && lsub == 3) ? kTRUE : kFALSE;
      }
   }

   Bool_t ChannelTree::GetIfoSubLoc (const char* chn, 
                     char* ifo, char* sub, char* loc, char* rest)
   {
      *ifo = 0;
      *sub = 0;
      *loc = 0;

      const char* cur = chn;

      //  find the ':', copy preceding text to ifo;
      const char* p = strchr (cur, ':');
      if (p == 0) return kFALSE;
      size_t lifo = p - cur;
      if (lifo > 63) lifo = 63;
      memcpy(ifo, cur, lifo);
      ifo[lifo] = 0;
      cur = p + 1;

      // vacuum channels!
      char* w = strchr (ifo, '-');
      if (w) {
	 //  Move subsystem name from ifo
	 size_t lsub = lifo;
	 lifo = w - ifo;
	 *w++ = 0;

	 lsub -= lifo + 1;
         memcpy(sub, w, lsub);
         sub[lsub] = 0;

	 //  Find separator between location and rest.
         p = strchr (cur, '_');
         if (p == 0) {
            strcpy (rest, "*");
	    size_t lrest = strlen(cur);
	    if (lrest > 255) lrest = 255;
            memcpy(loc, cur, lrest);
            rest[lrest] = 0;
         }
         else {
	    size_t len = p - cur;
	    if (len > 255) len = 255;
	    strncpy (loc, cur, len);
	    loc[len] = 0;

	    len = strlen(++p);
	    if (len > 255) len = 255;
            memcpy(rest, p, len);
            rest[len] = 0;
	 }
         return  (lifo == 3 && lsub == 2) ? kTRUE : kFALSE;
      }
      // normal channels
      else {
	 // Copy subsystem name
         p = strchr (cur, '-');
         if (p == 0) {
            return kFALSE;
         }
	 size_t lsub = p - cur;
	 if (lsub > 63) lsub = 63;
         memcpy (sub, cur, lsub);
         sub[lsub] = 0;
         cur = p + 1;

	 // Copy rest if available
         p = strchr (cur, '_');
         if (p == 0) {
            strcpy (rest, "*");
	    size_t lrest = strlen(cur);
	    if (lrest > 255) lrest = 255; 
	    memcpy(loc, cur, lrest);
	    loc[lrest] = 0;
         }
         else {
	    size_t len = p - cur;
	    if (len > 255) len = 255; 
	    memcpy(loc, cur, len);
	    loc[len] = 0;

	    len = strlen(++p);
	    if (len > 255) len = 255; 
            strncpy (rest, p, len);
            rest[len] = 0;
         }

	 // Copy location if present
         return (lifo == 2 && lsub == 3) ? kTRUE : kFALSE;
      }
   }

extern "C" 
   int ChannelTree_channelcmp (const void* s1, const void* s2) {
      return (strcasecmp (((ChannelEntry*)s1)->Name(), 
                         ((ChannelEntry*)s2)->Name()));
   }

extern "C" 
   int ChannelTree_channelcmprate (const void* s1, const void* s2) {
      const ChannelEntry* c1 = (ChannelEntry*)s1;
      const ChannelEntry* c2 = (ChannelEntry*)s2;
      if (((c1->Rate() > 0) && (c1->Rate() <= 16)) ^
         ((c2->Rate() > 0) && (c2->Rate() <= 16))) {
         return ((c1->Rate() > 0) && (c1->Rate() <= 16)) ? 1 : -1;
      }
      else {
         return strcasecmp (c1->Name(), c2->Name());
      }
   }

   static TString displayedName (const char* rest, Int_t type, 
				 const ChannelEntry* chn)
   {
      char buf[512];
      int mode = 0;
      if ((type & kChannelTreeShowRate)   &&   chn->Rate() > 0)  mode += 1;
      if ((type & kChannelTreeShowSource) && *(chn->UDN()) != 0) mode += 2;
      switch (mode) {
      case 0:
	 sprintf(buf, "%s", rest);
	 break;
      case 1:
	 sprintf(buf, "%s (%g)", rest, chn->Rate());
	 break;
      case 2:
	 sprintf(buf, "%s ( @ %s)", rest, chn->UDN());
	 break;
      case 3:
	 sprintf(buf, "%s (%g @ %s)", rest, chn->Rate(), chn->UDN());
	 break;
      }
      return TString(buf);
   }

   //--------  Build the channel tree
   // The channel tree is displayed in channel selection combo box widgets
   // as a tree with branches for IFO, subsystem, and location.  As an example
   // if a channel name is H2:SUS-ITMY_BQF_EXC, the IFO is H2, the subsystem is
   // SUS, and the location is ITMY.  
   Bool_t ChannelTree::BuildChannelTree ()
   {
      // Get rid of any old data for the tree.
      DeleteChannels();
      if (fChannels == 0) {
         return kTRUE;
      }
   
      // sort first - The list must be sorted so we can easily put
      // channels into the branches of the tree.
      if (fChannelTreeType & kChannelTreeSeparateSlow) {
         ::qsort (fChannels, fChnNum, sizeof (ChannelEntry), 
                 ChannelTree_channelcmprate);
      }
      else {
         ::qsort (fChannels, fChnNum, sizeof (ChannelEntry), 
                 ChannelTree_channelcmp);
      }
   
      // go through list and add to tree
      char prevIfo[64] = {0};
      char ifo[64] = {0};
      TLGLBTreeEntry* prevIfoItems = 0;
      char prevSub[64] = {0};
      char sub[64] = {0};
      TLGLBTreeEntry* prevSubItems = 0;
      char prevLoc[256] = {0};
      char loc[256] = {0};
      TLGLBTreeEntry* prevLocItems = 0;
      char rest[256] = {0};
      int* oddItems = new int [fChnNum];
      int oddNum = 0;
      if (oddItems == 0) {
         return kFALSE;
      }
   
      for (UInt_t i = 0; i < fChnNum; i++) {
         bool succ = false;
         if (fChannelTreeType & kChannelTreeLevel3) {
	     // Parse the channel name into IFO, subsystem, and location.
            succ = GetIfoSubLoc (fChannels[i].Name(), ifo, sub, loc, rest);
         }
         else {
	    // Parse the channel name into IFO and subsystem.
            succ = GetIfoSub (fChannels[i].Name(), ifo, sub, rest);
         }
         if (succ) {
	    // Use the data rate to decide if this is a slow channel or fast.
            if (fChannelTreeType & kChannelTreeSeparateSlow) {
               if ((fChannels[i].Rate() > 0) && 
                  (fChannels[i].Rate() <= 16)) {
                  strcpy (ifo + strlen (ifo), " (slow)");
               }
            }
            // create site/ifo branch if necessary
            if ((strcasecmp (prevIfo, ifo) != 0) ||
               (prevIfoItems == 0)) {
               prevIfoItems = AddChannel (0, ifo);
               strcpy (prevIfo, ifo);
	       // Each time a new ifo branch is created, a new subsystem branch must be started.
               prevSubItems = 0;
               strcpy (prevSub, "");
            }
            if (prevIfoItems == 0) {
               delete [] oddItems;
               return kFALSE;
            }
            // create subsystem branch if necessary
            if ((strcasecmp (prevSub, sub) != 0) ||
               (prevSubItems == 0)) {
               prevSubItems = AddChannel (prevIfoItems, sub);
               strcpy (prevSub, sub);
	       // Each time a new subsystem branch is created, a new location branch must be started.
	       prevLocItems = 0 ;	// JCB bugzilla 356
	       strcpy (prevLoc, "") ;	// JCB bugzilla 356
            }
            if (prevSubItems == 0) {
               delete [] oddItems;
               return kFALSE;
            }
            // create location branch if necessary
            if (fChannelTreeType & kChannelTreeLevel3) {
               if ((strcasecmp (prevLoc, loc) != 0) ||
                  (prevLocItems == 0)) {
                  prevLocItems = AddChannel (prevSubItems, loc);
                  strcpy (prevLoc, loc);
               }
               if (prevLocItems == 0) {
                  delete [] oddItems;
                  return kFALSE;
               }
            }
            // create channel entry
            TLGLBTreeEntry* pItems = (fChannelTreeType & kChannelTreeLevel3) ?
               prevLocItems : prevSubItems;
            TString entryname = 
               displayedName (rest, fChannelTreeType, fChannels + i);
            if (fChnClient == 0) {
               AddChannel (pItems, entryname, fChannels[i].Name(), i);
            }
            else {
               AddChannel (pItems, entryname, fChannels[i].Name(), i,
                          fLeftArrow, fLeftArrow);
            }
         }
         // oddball channel names: mark them and add at end
         else {
            oddItems[oddNum] = i;
            oddNum++;
         }
      }
   
      // check odd channel names
      if (oddNum > 0) {
         prevIfoItems = AddChannel (0, "others");
         if (prevIfoItems == 0) {
            delete [] oddItems;
            return kFALSE;
         }
         for (int i = 0; i < oddNum; i++) {
            TString entryname = 
               displayedName (fChannels[oddItems[i]].Name(), 
                             fChannelTreeType, fChannels + oddItems[i]);
            if (fChnClient == 0) {
               AddChannel (prevIfoItems, entryname,
                          fChannels[oddItems[i]].Name(), oddItems[i]);
            }
            else {
               AddChannel (prevIfoItems, entryname, 
                          fChannels[oddItems[i]].Name(), oddItems[i], 
                          fLeftArrow, fLeftArrow);
            }
         }
      }
   
      delete [] oddItems;
      return kTRUE;
   }


   Int_t ChannelTree::GetChannelId (const char* channelname) const 
   {
      for (UInt_t i = 0; i < fChnNum; i++) {
         if (strcasecmp (fChannels[i].Name(), channelname) == 0) {
            return i;
         }
      }
      return -1;
   }


   const char* ChannelTree::GetChannelName (Int_t id) const
   {
      if ((id < 0) || ((UInt_t)id >= fChnNum)) {
         return 0;
      }
      else {
         return fChannels[id].Name();
      }
   }



//////////////////////////////////////////////////////////////////////////
//                                                                      //
// TLGChannelListbox                                                    //
//                                                                      //
// A tree list box to deal with channel names                           //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

   TLGChannelListbox::TLGChannelListbox (TGWindow* p, Int_t id, 
                     ChannelEntry* chnnames, UInt_t chnnum, 
                     Int_t type, UInt_t option, ULong_t back)
   :  TLGLBTree (p, id, option, back), 
   ChannelTree (chnnames, chnnum, type)
   {
      SetChannelClient(fClient);
      BuildChannelTree();
   }


   TLGChannelListbox::TLGChannelListbox (TGWindow* p, Int_t id, 
                     const char* chnnames, Int_t type, 
                     UInt_t option, ULong_t back)
   : TLGLBTree (p, id, option, back), ChannelTree (chnnames, type)
   {
      SetChannelClient(fClient);
      BuildChannelTree();
   }


   TLGChannelListbox::~TLGChannelListbox ()
   {
   }



//////////////////////////////////////////////////////////////////////////
//                                                                      //
// TLGChannelCombobox                                                   //
//                                                                      //
// A tree combobox to deal with channel names                           //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

   TLGChannelCombobox::TLGChannelCombobox (TGWindow* p, Int_t id, 
                     ChannelEntry* chnnames, UInt_t chnnum,
                     Int_t type, Bool_t editable, 
                     UInt_t option, ULong_t back)
   : TLGComboTree (p, id, editable, option),
     ChannelTree (chnnames, chnnum, type)
   {
      SetChannelClient(fClient);
      BuildChannelTree();
   }


   TLGChannelCombobox::TLGChannelCombobox (TGWindow* p, Int_t id, 
                     const char* chnnames, Int_t type, 
                     Bool_t editable, UInt_t option, ULong_t back)
   : TLGComboTree (p, id, editable, option), ChannelTree (chnnames, type)
   {
      SetChannelClient(fClient);
      BuildChannelTree();
   }


   TLGChannelCombobox::~TLGChannelCombobox ()
   {
   }


}
